(*
 Irina Kurbatova
 Voronezh State University,
 Department of Applied Mathematics and Mechanics,
 Universitetskaya pl., 1,
 394018 Voronezh, Russia
 E-mail: la_soleil@bk.ru
*)

BeginPackage["LinearRelations`", {"LinearAlgebra`MatrixManipulation`"}]; 
Unprotect[Addition,HarmonicAddition,Multiplication,InverseRelation,IdentityRelation,ZeroRelation,InfiniteRelation,TotalRelation,EmptyRelation,MatrixToRelation,MatrixFormRelation,ColomnReduceRelation,Section,ImageOfZeroRelation,NullSpaceRelation,ConjugateRelation]; 

IdentityRelation::usage = "IdentityRelation[n] gives the relation that correspons to the identity matrix."; 
ZeroRelation::usage = "ZeroRelation[m,k] gives the zero relation Theta:\!\(R\^m\)->\!\(R\^k\). The zero relation is defined everywhere and takes all vectors to zero."; 
InfiniteRelation::usage = "InfiniteRelation[m,k] gives the infinite relation Omega:\!\(R\^m\)->\!\(R\^k\). The infinite relation is defined only at zero and takes the zero to all vectors."; 
EmptyRelation::usage = "EmptyRelation[m,k] gives the empty relation 0\[Cross]0:\!\(R\^m\)->\!\(R\^k\). The empty relation is defined only at zero and takes the zero to zero."; 
TotalRelation::usage = "TotalRelation[m,k] gives the total relation \!\(R\^m\)\[Cross]\!\(R\^k\):\!\(R\^m\)->\!\(R\^k\). The total relation is defined everywhere and takes all vectors to all vectors. Uses the package LinearAlgebra`MatrixManipulation`."; 
InverseRelation::usage = "InverseRelation[a] gives the inverse relation to the relation a."; 
ColomnReduceRelation::usage = "ColomnReduceRelation[a] gives the row\[Hyphen]reduced form of the relation a.";  
MatrixToRelation::usage = "MatrixToRelation[a] represents the matrix a in the form of relation."; 
MatrixFormRelation::usage = "MatrixFormRelation[a] shows the relation in the form of two matrices."; 
Section::usage = "Section[a] gives the section of the relation a.";  
ImageOfZeroRelation::usage = "ImageOfZeroRelation[a] gives the image of zero of the relation a.";  
NullSpaceRelation::usage = "NullSpaceRelation[a] gives the null space of the relation a.";  
ConjugateRelation::usage = "ConjugateRelation[a] gives the cojugate of the relation a.";  

Addition::usage = "Addition[a,b] gives the sum A+B of the two relations A:\!\(R\^m\)->\!\(R\^k\) and B:\!\(R\^m\)->\!\(R\^k\). 
It is assumed that A is defined by the list {a1,a2} of matrices and B is defined by  the list {b1,b2} of matrices. 
Uses the package LinearAlgebra`MatrixManipulation`."; 
HarmonicAddition::usage = "HarmonicAddition[a,b] gives the harmonic sum A+B of the two relations A:\!\(R\^m\)->\!\(R\^k\) and B:\!\(R\^m\)->\!\(R\^k\). 
It is assumed that A is defined by the list {a1,a2} of matrices and B is defined by the list {b1,b2} of matrices. The harmonic sum of A and B is \!\(\((A\^\(-1\) + B\^\(-1\))\)\^\(-1\)\).
Uses the package LinearAlgebra`MatrixManipulation`."; 
Multiplication::usage = "Multiplication[a,b] gives the product AB of the two relations A:\!\(R\^m\)->\!\(R\^k\) and B:\!\(R\^l\)->\!\(R\^m\). 
It is assumed that A is defined by the list {a1,a2} of matrices and B is defined by the list {b1,b2} of matrices. 
Uses the package LinearAlgebra`MatrixManipulation`."; 

Begin["`Private`"]; 

IdentityRelation[m_]:={IdentityMatrix[m],IdentityMatrix[m]};
ZeroRelation[m_,k_]:={IdentityMatrix[m],Table[0,{k},{m}]};
InfiniteRelation[m_,k_]:={Table[0,{m},{k}],IdentityMatrix[k]};
EmptyRelation[m_,k_]:={Table[{0},{m}],Table[{0},{k}]};
TotalRelation[m_,k_]:={BlockMatrix[{{IdentityMatrix[m],ZeroMatrix[m,k]}}], BlockMatrix[{{ZeroMatrix[k,m],IdentityMatrix[k]}}]};      
InverseRelation[a_]:={a[[2]],a[[1]]};
ColomnReduceRelation[a_]:=Module[{m=Length[a[[1]]],k=Length[a[[2]]],n=Length[a[[1,1]]],answer},
answer=BlockMatrix[{{Transpose[a[[1]]],Transpose[a[[2]]]}}];
answer=RowReduce[answer];
While[And[n>1,Sum[Abs[answer[[n,j]]],{j,m+k}]\[Equal]0],answer=Delete[answer,n];n--];
answer=Transpose[answer];
answer={TakeRows[answer,{1,m}],TakeRows[answer,{m+1,m+k}]}];
MatrixToRelation[a_]:=Module[{k=Length[a[[1]]]},{IdentityMatrix[k],a}];
MatrixFormRelation[a_]:=MatrixForm[{MatrixForm[a[[1]]],MatrixForm[a[[2]]]}];
Section[a_]:=Module[{m=Length[a[[1]]],k=Length[a[[2]]],b=ColomnReduceRelation[a],a1,a2,n,answer={}},
a1=Transpose[b[[1]]];a2=Transpose[b[[2]]];
n=Length[a1];  
While[And[n>0,Sum[Abs[a1[[n,j]]],{j,m}]\[Equal]0],answer=Prepend[answer,a2[[n]]];a1=Delete[a1,n];a2=Delete[a2,n];n--];
{Transpose[a1],Transpose[a2]}];
ImageOfZeroRelation[a_]:=Module[{m=Length[a[[1]]],k=Length[a[[2]]],b=ColomnReduceRelation[a],a1,a2,n,answer={}},
a1=Transpose[b[[1]]];a2=Transpose[b[[2]]];
n=Length[a1];
While[And[n>0,Sum[Abs[a1[[n,j]]],{j,m}]\[Equal]0],answer=Prepend[answer,a2[[n]]];a1=Delete[a1,n];a2=Delete[a2,n];n--];
If[answer=={},EmptyRelation[m,k],
  answer={Table[0,{m},{Length[answer]}],Transpose[answer]}]];
NullSpaceRelation[a_]:=InverseRelation[ImageOfZeroRelation[InverseRelation[a]]];
ConjugateRelation[a_]:=Module[{m=Length[a[[1]]],k=Length[a[[2]]],e},
e=NullSpace[BlockMatrix[{{Transpose[a[[1]]],-Transpose[a[[2]]]}}]];
e=If[e\[Equal]{},Table[{0},{m+k}],Transpose[e]];
{TakeRows[e,{m+1,m+k}],TakeRows[e,{1,m}]}];

Addition[a_,b_]:=Module[{ans1={},ans2={},ans,answer={{},{}},e,nE,
nA1=Length[a[[1,1]]],nA2=Length[a[[2,1]]],nA,
nB1=Length[b[[1,1]]],nB2=Length[b[[2,1]]],nB,
ma=Length[a[[1]]],mb=Length[b[[1]]],m,
ka=Length[a[[2]]],kb=Length[b[[2]]],k},
Addition::"dim"="Tensors `1` and `2` have incompatible shapes.";
If[ma\[Equal]mb,m=ma,Message[Addition::"dim",a[[1]],b[[1]]]; Goto[finish]];
If[ka\[Equal]kb,k=ka,Message[Addition::"dim",a[[2]],b[[2]]];Goto[finish]];
If[nA1\[Equal]nA2,nA=nA1,Message[Addition::"dim",a[[1]],a[[2]]];Goto[finish]];
If[nB1\[Equal]nB2,nB=nB1,Message[Addition::"dim",b[[1]],b[[2]]];Goto[finish]];
e=NullSpace[BlockMatrix[{{a[[1]],-b[[1]]}}]];
If[e\[Equal]{},answer=EmptyRelation[m,k],
nE=Length[e];
wB=Table[Null,{k}];wA=Table[Null,{m}];
Do[Do[wA[[i]]=Sum[e[[kk,j]]*a[[1,i,j]],{j,nA}],{i,m}];
   Do[wB[[i]]=Sum[e[[kk,j]]*a[[2,i,j]],{j,nA}]+
              Sum[e[[kk,nA+j]]*b[[2,i,j]],{j,nB}],{i,k}];
   ans1=Append[ans1,wA];
   ans2=Append[ans2,wB],
   {kk,nE}];
answer={Transpose[ans1],Transpose[ans2]}];
Label[finish];
Chop[answer]];
HarmonicAddition[a_,b_]:=InverseRelation[Addition[InverseRelation[a],InverseRelation[b]]];
Multiplication[a_,b_]:=Module[{ansA={},ansB={},answer={{},{}},e,nE,
nA1=Length[a[[1,1]]],nA2=Length[a[[2,1]]],nA,
nB1=Length[b[[1,1]]],nB2=Length[b[[2,1]]],nB,
l=Length[b[[1]]],ma=Length[a[[1]]],mb=Length[b[[2]]],m,k=Length[a[[2]]]},
Multiplication::"dim"="Tensors `1` and `2` have incompatible shapes.";
If[ma\[Equal]mb,m=ma,Message[Multiplication::"dim",a[[1]],b[[2]]];Goto[finish]];
If[nA1\[Equal]nA2,nA=nA1,Message[Multiplication::"dim",a[[1]],a[[2]]];Goto[finish]];
If[nB1\[Equal]nB2,nB=nB1,Message[Multiplication::"dim",b[[1]],b[[2]]];Goto[finish]];
e=NullSpace[BlockMatrix[{{a[[1]],-b[[2]]}}]];
If[e=={},answer=EmptyRelation[l,k],
nE=Length[e];
wB=Table[Null,{l}];wA=Table[Null,{k}];
Do[Do[wB[[j]]=Sum[e[[kk,nA+i]]*b[[1,j,i]],{i,nB}],{j,l}];
   Do[wA[[j]]=Sum[e[[kk,i]]*a[[2,j,i]],{i,nA}],{j,k}];
   ansA=Append[ansA,wA]; 
   ansB=Append[ansB,wB],
   {kk,nE}];
answer={Transpose[ansB],Transpose[ansA]}];
Label[finish];
Chop[answer]];

End[]; 
Protect[Addition,HarmonicAddition,Multiplication,InverseRelation,IdentityRelation,ZeroRelation,InfiniteRelation,TotalRelation,EmptyRelation,MatrixToRelation,MatrixFormRelation,ColomnReduceRelation,Section,ImageOfZeroRelation,NullSpaceRelation,ConjugateRelation]; EndPackage[]